import os
import re
import datetime
import collections

from . import rendering
from . import consts
from . import utils


DATESLUG_RE = re.compile(
    r"^(?P<date_string>\d\d\d\d-\d\d-\d\d)-(?P<slug>.+)$")


Network = collections.namedtuple(
    "Network",
    [
        "slug", "sites_slugs", "sites", "network_dir", "chunks_data",
        "translations_data", "logo_path"])
Site = collections.namedtuple(
    "Site",
    [
        "attrs", "slug", "notes", "series", "renderer",
        "network", "site_dir", "chunks_data", "tags", "logo_path",
        "earliest_year", "current_year", "translations_data"])
Note = collections.namedtuple(
    "Note",
    [
        "attrs", "body", "slug", "date", "tags_slugs", "series_slugs",
        "substatic_root", "substatic_suffixes"])
Tag = collections.namedtuple(
    "Tag", ["attrs", "description", "slug", "notes_slugs"])
Series = collections.namedtuple(
    "Series", ["attrs", "description", "slug", "notes_slugs"])


def prepare_sites(
        sites_dirs, networks_dir, internal_templates_dir,
        prepended_templates_dirs, appended_templates_dirs,
        current_date, error_callback, attr_overrides):
    attr_overrides = _parse_attr_overrides(attr_overrides)
    networks = _get_networks(networks_dir, error_callback=error_callback)
    # Make {site_slug: network.network_dir} mapping to make template
    # search path (to make renderer).
    _site_slug_to_network_dir = {
        site_slug: network.network_dir
        for network in networks
        for site_slug in network.sites_slugs}
    sites = []
    for site_dir in sites_dirs:
        site = _make_site(site_dir)
        internal_templates_dirs = [
            os.path.join(_site_slug_to_network_dir[site.slug], "templates"),
            os.path.join(site.site_dir, "templates"),
            internal_templates_dir]
        renderer = rendering.make_renderer(
            charset=consts.CHARSET, prepended=prepended_templates_dirs,
            internal=internal_templates_dirs,
            appended=appended_templates_dirs)
        # This needs explanation. We did _make_site(site_dir) to create object
        # with slug and site_dir fields, which we used to make template search
        # path. Now all other fields need to be filled (notes, tags, etc.)
        sites.append(
            _get_site(
                site, current_date=current_date, renderer=renderer,
                error_callback=error_callback,
                attr_override=attr_overrides.get(site.slug, {})))
    return _set_up_networks(networks=networks, sites=sites)


def _parse_attr_overrides(raw_overrides):
    attr_overrides = {}
    for s in raw_overrides:
        slug, file_path = s.split("=", 1)
        attr_overrides[slug] = utils.read_kv(file_path)
    return attr_overrides


def _get_networks(networks_dir, error_callback):
    def _get_network(network_dir, error_callback):
        network = Network(
            slug=os.path.basename(network_dir),
            sites_slugs=[], sites=[],
            network_dir=network_dir, chunks_data={}, translations_data={},
            logo_path=_make_logo_path(network_dir))
        network.sites_slugs.extend(
            utils.read_slugs(network_dir, "sites-slugs"))
        if not network.sites_slugs:
            error_callback(
                "{!r} network has less than one site!".format(network.slug))
        network.chunks_data.update(utils.read_chunks(network_dir))
        network.translations_data.update(utils.read_translations(network_dir))
        return network
    networks = []
    for network_dir in utils.subpaths(networks_dir):
        networks.append(
            _get_network(network_dir, error_callback=error_callback))
    return networks


def _make_site(site_dir):
    return Site(
        attrs={}, slug=os.path.basename(site_dir), notes=[], series={},
        renderer=None, network=None, site_dir=site_dir, chunks_data={},
        tags={}, logo_path=None, earliest_year=None, current_year=None,
        translations_data={})


def _get_site(site, current_date, renderer, error_callback, attr_override):
    site = site._replace(renderer=renderer)
    # Read translations.
    site.translations_data.update(
        utils.read_translations(site.site_dir))
    # Read attrs.
    site = site._replace(attrs=utils.read_attrs(site.site_dir))
    site.attrs.update(attr_override)
    utils.check_required_attrs(
        site.attrs, ("domain", "name", "date_format", "language"),
        entity_slug=site.slug,
        error_callback=error_callback)
    # Read tags.
    tags_dir = os.path.join(site.site_dir, "tags")
    for tag_dir in utils.subpaths(tags_dir):
        tag = _get_tag(
            tag_dir, renderer=renderer, error_callback=error_callback)
        site.tags[tag.slug] = tag
    # Read notes' series.
    all_series_dir = os.path.join(site.site_dir, "series")
    for series_dir in utils.subpaths(all_series_dir, ignore_not_exists=True):
        series = _get_series(
            series_dir, renderer=renderer, error_callback=error_callback)
        site.series[series.slug] = series
    # Read notes & populate site's notes list.
    notes_dir = os.path.join(site.site_dir, "notes")
    for note_dir in utils.subpaths(notes_dir):
        site.notes.append(
            _get_note(
                note_dir, tags=site.tags.values(),
                series=site.series.values(), renderer=renderer,
                error_callback=error_callback))
    # Set earliest year (based on publication dates of notes)
    # and current year (value passed).
    site = site._replace(
        earliest_year=min(note.date for note in site.notes).year,
        current_year=current_date.year)
    # Read chunks & populate site's chunks mapping.
    site.chunks_data.update(utils.read_chunks(site.site_dir))
    # Set logo path.
    site = site._replace(logo_path=_make_logo_path(site.site_dir))
    # Check for note slug duplication.
    dupes_counter = collections.Counter(note.slug for note in site.notes)
    for slug, count in dupes_counter.items():
        if count > 1:
            error_callback(
                "{!r} slug is used {} times in {!r} site!".format(
                    slug, count, site.slug))
    # Check notes do not reference non-existing tags.
    for note in site.notes:
        if not note.tags_slugs:
            error_callback(
                "{!r} note of {!r} site does not have tags!".format(
                    note.slug, site.slug))
        else:
            for slug in note.tags_slugs:
                if slug not in site.tags:
                    error_callback(
                        (
                            "{!r} tag is undefined but is used "
                            "by {!r} note of {!r} site!").format(
                                slug, note.slug, site.slug))
    # Check tags have at least one associated note.
    for slug, tag in site.tags.items():
        if not tag.notes_slugs:
            error_callback(
                (
                    "{!r} tag of {!r} site does not have "
                    "any notes associated!").format(
                        slug, site.slug))
    # Check series have at least one associated note.
    for slug, series in site.series.items():
        if not series.notes_slugs:
            error_callback(
                (
                    "{!r} series of {!r} site does not have "
                    "any notes associated!").format(
                        slug, site.slug))
    return site


def _get_note(note_dir, tags, series, renderer, error_callback):
    def _get_date_and_slug(note_dir):
        # '.../2015-08-08-some-slug' -> (date(), 'some-slug')
        dateslug = os.path.basename(note_dir)
        matchobj = DATESLUG_RE.match(dateslug)
        date_string = matchobj.group("date_string")
        date = datetime.datetime.strptime(date_string, "%Y-%m-%d").date()
        return date, matchobj.group("slug")
    date, slug = _get_date_and_slug(note_dir)
    attrs = utils.read_attrs(note_dir)
    utils.check_required_attrs(
        attrs, ("title", ),
        entity_slug=slug,
        error_callback=error_callback)
    body = renderer.render_markdown(utils.read_subfile(note_dir, "body.md"))
    tags_slugs = {tag.slug for tag in tags if slug in tag.notes_slugs}
    series_slugs = {s.slug for s in series if slug in s.notes_slugs}
    substatic_root, substatic_suffixes = utils.get_substatic_data(note_dir)
    return Note(
        slug=slug, date=date, body=body, attrs=attrs, tags_slugs=tags_slugs,
        series_slugs=series_slugs, substatic_root=substatic_root,
        substatic_suffixes=substatic_suffixes)


def _get_tag(tag_dir, renderer, error_callback):
    slug = os.path.basename(tag_dir)
    attrs = utils.read_attrs(tag_dir)
    utils.check_required_attrs(
        attrs, ("name", ),
        entity_slug=slug,
        error_callback=error_callback)
    try:
        description = renderer.render_text(
            utils.read_subfile(tag_dir, "description"))
    except IOError:
        description = None
    notes_slugs = set(utils.read_slugs(tag_dir, "notes-slugs"))
    return Tag(
        attrs=attrs, description=description, slug=slug,
        notes_slugs=notes_slugs)


def _get_series(series_dir, renderer, error_callback):
    slug = os.path.basename(series_dir)
    attrs = utils.read_attrs(series_dir)
    utils.check_required_attrs(
        attrs, ("name", ),
        entity_slug=slug,
        error_callback=error_callback)
    try:
        description = renderer.render_text(
            utils.read_subfile(series_dir, "description"))
    except IOError:
        description = None
    notes_slugs = set(utils.read_slugs(series_dir, "notes-slugs"))
    return Series(
        attrs=attrs, description=description, slug=slug,
        notes_slugs=notes_slugs)


def _set_up_networks(networks, sites):
    slug_to_site = {site.slug: site for site in sites}
    site_slug_to_network = {}
    for unfilled_network in networks:
        slugs = unfilled_network.sites_slugs
        network = unfilled_network._replace(
            sites=[slug_to_site[slug] for slug in slugs])
        for slug in slugs:
            site_slug_to_network[slug] = network
    return list(
        site._replace(network=site_slug_to_network[site.slug])
        for site in sites)


def _make_logo_path(base_dir):
    return os.path.join(base_dir, "logo.xcf")



import os
import collections

from . import errorpages
from . import utils


_TEXT_CONTENT_TYPES = (
    "text/javascript",
    "application/javascript",
    "text/css",
    "application/atom+xml")
# ^ "text/html" is missing because it is added by nginx by default.
_BINARY_CONTENT_TYPES = ("image/x-icon", )
GZIPPED_CONTENT_TYPES = " ".join(
    (_TEXT_CONTENT_TYPES + _BINARY_CONTENT_TYPES))
CHARSET_CONTENT_TYPES = " ".join((_TEXT_CONTENT_TYPES))


Directive = collections.namedtuple(
    "Directive", ["name", "arg", "children", "level"])


def _get_servers(specs):
    def _end_slash(path):
        if not path.endswith("/"):
            return "".join((path, "/"))
        return path

    def mk(name, arg="", children=(), level=1):
        def _set_level(children, current_level):
            for ch in children:
                yield ch._replace(
                    level=current_level + 1,
                    children=list(_set_level(ch.children, current_level + 1)))
        children = list(_set_level(children, level))
        return Directive(name=name, arg=arg, children=children, level=level)

    for spec in sorted(specs, key=lambda spec: spec.site.slug):
        domain = spec.site.attrs["domain"]
        yield (
            mk("listen", "80"),
            mk("server_name", "www.{}".format(domain)),
            mk("return", "301 $scheme://{}$request_uri".format(domain)))
        yield (
            mk("listen", "80"),
            mk("server_name", domain),
            mk("root", spec.site_build_dir),
            mk("client_header_buffer_size", "1k"),
            mk("large_client_header_buffers", "4 8k"),
            mk("client_max_body_size", "1M"),
            mk("gzip", "on"),
            mk("gzip_disable", "msie6"),
            mk("gzip_types", GZIPPED_CONTENT_TYPES),
            mk("gzip_vary", "on"),
            mk("gzip_comp_level", "6"),
            mk("charset", "utf-8"),
            mk("source_charset", "utf-8"),
            mk("charset_types", CHARSET_CONTENT_TYPES),
            mk("error_log", "/var/log/nginx/error.log crit"),
            mk("access_log", "off"),
            mk(
                "location", "/", (
                    mk("root", spec.pages_build_dir),
                    mk("index", "index.html index.atom"),
                    mk("expires", "1d"),
                    mk("break"))),
            mk(
                "location", "/s/", (
                    mk("alias", _end_slash(spec.static_build_dir)),
                    mk("expires", "1d"),
                    mk("break"))),
            ) + tuple(
                mk(
                    "error_page",
                    "{} {}".format(code, errorpages.make_url_path(code)))
                for code in sorted(errorpages.CODES))


def _get_lines(directive):
    wsp = directive.level * "    "
    tmpl_prefix = "{wsp}{d.name}"
    if directive.arg:
        tmpl_prefix += " {d.arg}"
    if directive.children:
        yield (tmpl_prefix + " {{").format(d=directive, wsp=wsp)
        for child in directive.children:
            yield from _get_lines(child)
        yield "{wsp}}}".format(wsp=wsp)
    else:
        yield (tmpl_prefix + ";").format(d=directive, wsp=wsp)


def build_nginx_config(specs, build_dir):
    nginx_config_path = os.path.join(build_dir, "etc", "nginx.conf")
    os.makedirs(os.path.dirname(nginx_config_path), exist_ok=True)
    lines = []
    for server in _get_servers(specs):
        lines.append("server {")
        for directive in server:
            lines.extend(_get_lines(directive))
        lines.append("}\n")
    utils.write_file(nginx_config_path, "\n".join(lines))



from __future__ import print_function

import os
import sys
import locale
import fnmatch
import datetime
import argparse
import collections

from paka.vx1.preparation import prepare_sites
from paka.vx1.building import prepare_build_dir, build_site_by_spec
from paka.vx1.static import build_static
from paka.vx1.icons import build_icons
from paka.vx1.nginx import build_nginx_config
from paka.vx1.features import make_feature_checker
from paka.vx1.utils import subpaths


INTERNAL_TEMPLATES_DIR = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "templates"))


SiteBuildSpec = collections.namedtuple(
    "SiteBuildSpec",
    ["site", "site_build_dir", "pages_build_dir", "static_build_dir"])


def _set_build_dir(spec, build_dir):
    site_build_dir = os.path.join(build_dir, "sites", spec.site.slug)
    return spec._replace(
        site_build_dir=site_build_dir,
        pages_build_dir=os.path.join(site_build_dir, "pages"),
        static_build_dir=os.path.join(site_build_dir, "s"))


def set_build_dir(specs, build_dir):
    return list(_set_build_dir(spec, build_dir=build_dir) for spec in specs)


def make_sites_specs(sites, build_dir):
    def mk(site):
        return SiteBuildSpec(
            site=site, site_build_dir=None, pages_build_dir=None,
            static_build_dir=None)
    return set_build_dir((mk(site) for site in sites), build_dir=build_dir)


def error_callback(text):
    print("[ERROR] {}".format(text), file=sys.stderr)
    sys.exit(1)


def main(argv=sys.argv[1:]):
    locale.setlocale(locale.LC_ALL, "C")
    parser = argparse.ArgumentParser(prog="paka.vx1")
    parser.add_argument(
        "--blognets-dir",
        help="dir containing input data (sites and networks)",
        required=True)
    parser.add_argument(
        "--build-dir", help="dir that will contain output data",
        required=True)
    parser.add_argument(
        "--build-dir-future-path",
        help=(
            "path at which build dir will be on prod "
            "(to generate correct config for nginx)"))
    parser.add_argument(
        "--template-dirs-prepend", nargs="*",
        help=(
            "paths to dirs with templates to prepend to template search path"))
    parser.add_argument(
        "--template-dirs-append", nargs="*",
        help=(
            "paths to dirs with templates to append to template search path"))
    parser.add_argument(
        "--slug-pattern", required=True, default="*",
        help="Unix shell-style pattern for slugs of sites to build")
    parser.add_argument("--cache-dir", required=True, help="dir for cache")
    parser.add_argument(
        "--current-date",
        help="date (in %%Y-%%m-%%d format) to use as current")
    parser.add_argument(
        "--features", nargs="*", default=[],
        help="names of build features to turn on")
    parser.add_argument(
        "--site-attr-overrides", action="append",
        help="site_slug=path_to_attrs_file_with_pairs_to_update")
    args = parser.parse_args(argv)

    if args.current_date:
        current_date = datetime.datetime.strptime(
            args.current_date, "%Y-%m-%d").date()
    else:
        current_date = datetime.date.today()

    blognets_dir = os.path.abspath(args.blognets_dir)
    networks_dir = os.path.join(blognets_dir, "networks")
    sites_dir = os.path.join(blognets_dir, "sites")

    template_dirs_prepended = [
        os.path.abspath(path)
        for path in (args.template_dirs_prepend or ())]
    template_dirs_appended = [
        os.path.abspath(path)
        for path in (args.template_dirs_append or ())]

    build_dir = os.path.abspath(args.build_dir)
    build_dir_future_path = os.path.abspath(
        args.build_dir_future_path or build_dir)
    cache_dir = os.path.abspath(args.cache_dir)

    feature_checker = make_feature_checker(
        args.features, error_callback=error_callback)

    sites = prepare_sites(
        sites_dirs=subpaths(sites_dir),
        networks_dir=networks_dir,
        internal_templates_dir=INTERNAL_TEMPLATES_DIR,
        prepended_templates_dirs=template_dirs_prepended,
        appended_templates_dirs=template_dirs_appended,
        current_date=current_date, error_callback=error_callback,
        attr_overrides=args.site_attr_overrides or ())
    # Determine which sites to build.
    slug_to_site = {site.slug: site for site in sites}
    sites = [
        slug_to_site[slug]
        for slug in fnmatch.filter(slug_to_site, args.slug_pattern)]
    # Build sites.
    prepare_build_dir(build_dir)
    specs = make_sites_specs(sites, build_dir=build_dir)
    for spec in specs:
        build_site_by_spec(
            spec, feature_checker=feature_checker,
            error_callback=error_callback)
    # Build static (CSS, JS).
    build_static(specs)
    # Build favicon, etc.
    build_icons(
        specs, error_callback=error_callback,
        cache_dir=os.path.join(cache_dir, "icons"))
    # Build nginx config.
    specs = set_build_dir(specs, build_dir=build_dir_future_path)
    build_nginx_config(specs, build_dir=build_dir)



import os
import shutil
import hashlib
import subprocess


def build_icons(specs, error_callback, cache_dir):
    os.makedirs(cache_dir, exist_ok=True)
    for spec in specs:
        in_path = _get_existing_logo_file_path(
            spec.site, error_callback=error_callback)
        in_path_hash = _hash_file_contents(in_path, algorithm="sha1")
        cached_path = os.path.join(cache_dir, in_path_hash)
        if not os.path.exists(cached_path):
            os.mkdir(cached_path)
            for cmd in _make_cmds(in_path, cached_path):
                _call(cmd, error_callback=error_callback)
        for icon_file_name in os.listdir(cached_path):
            shutil.copy(
                os.path.join(cached_path, icon_file_name),
                spec.pages_build_dir)


def _get_existing_logo_file_path(site, error_callback):
    paths = (site.logo_path, site.network.logo_path)
    for path in paths:
        if os.path.isfile(path):
            return path
    error_callback(
        "logo file does not exist for {!r} site!".format(
            site.slug))


# Copied from storage/storage_library/checksumming.py.
def _iter_file_chunks(file_path, chunk_size=64 * 1024 * 1024):
    with open(file_path, "rb") as file:
        chunk = file.read(chunk_size)
        while chunk:
            yield chunk
            chunk = file.read(chunk_size)


# Copied from storage/storage_library/checksumming.py.
def _hash_file_contents(file_path, algorithm):
    h = hashlib.new(algorithm)
    for chunk in _iter_file_chunks(file_path):
        h.update(chunk)
    return h.hexdigest()


# Adapted from skel/priv/dolib/api.py.
def _call(args, error_callback):
    return_code = subprocess.call(args)
    if return_code:
        error_callback(
            "Subprocess error.\n{}".format(" ".join(args)))


# Adapted from skel/priv/plugins/static-for-web.py.
def _make_cmds(in_image_path, out_dir):
    def cmd(args, out_file_name, command="convert", unsharp=True):
        r = [command, in_image_path, "-layers", "merge"]
        r.extend(args)
        if unsharp:
            r.extend(["-unsharp", "0x1"])
        r.append(os.path.join(out_dir, out_file_name))
        return r
    # As icons are square, store just one side.
    apple_touch_icon_side_sizes = {
        57, 60, 72, 76, 114, 120, 144, 152, 180, 192}
    favicon_side_sizes = {16, 32, 48, 64}
    # 1. Icons for iOS and Android.
    for side_size in sorted(apple_touch_icon_side_sizes):
        s = "{side_size}x{side_size}".format(side_size=side_size)
        for basename_suffix in {"", "-precomposed"}:
            filename = "apple-touch-icon-{}{}.png".format(s, basename_suffix)
            yield cmd(["-resize", s], filename, unsharp=True)
    yield cmd(["-resize", "180x180"], "apple-touch-icon.png", unsharp=True)
    yield cmd(
        ["-resize", "180x180"], "apple-touch-icon-precomposed.png",
        unsharp=True)
    # 2. Favicon.
    size_args = []
    for side_size in sorted(favicon_side_sizes):
        s = "{side_size}x{side_size}".format(side_size=side_size)
        size_args.extend(["(", "-clone", "0", "-resize", s, ")"])
    args = (
        ["-border", "0"] +
        size_args +
        ["-delete", "0", "-alpha", "off", "-colors", "256"])
    yield cmd(args, "favicon.ico", unsharp=True)



import markupsafe


_DATA_CACHE = {}


def translate(key, context, site, escape=True):
    cache_key = "_".join((site.network.slug, site.slug))
    data = _DATA_CACHE.get(cache_key)
    if data is None:
        _DATA_CACHE[cache_key] = data = dict(
            site.network.translations_data, **site.translations_data)
    value = data[key].format(**context).strip()
    if escape:
        value = markupsafe.escape(value)
    return value



import os

import paka.webstatic.pipeline as _pipeline


def _make_output(spec, filename):
    return _pipeline.Output(
        os.path.join(spec.static_build_dir, filename),
        makedirs=True)


def _make_input(spec, template_name):
    return _pipeline.InputItem(
        None, data=spec.site.renderer(template_name))


def _gen_css_pipeline(spec):
    yield from (
        _make_input(spec, "styles.css.mako"),
        _pipeline.CSSMin(),
        _make_output(spec, "styles.css"))


def _gen_js_pipeline(spec):
    yield from (
        _make_input(spec, "scripts.js.mako"),
        _pipeline.JSMin(),
        _make_output(spec, "scripts.js"))


def build_static(specs):
    for spec in specs:
        for func in (_gen_css_pipeline, _gen_js_pipeline):
            _pipeline.run(func(spec))



import os

import markupsafe


CHARSET = "utf-8"


def write_file(path, contents):
    with open(path, "wb") as file:
        file.write(contents.encode(CHARSET))


def read_file(path):
    """Read file and return its contents as str."""
    with open(path, "rb") as file:
        binary_contents = file.read()
    return binary_contents.decode(CHARSET)


def read_subfile(base_dir, subfile_name):
    """Read file located in dir and return its contents as str."""
    return read_file(os.path.join(base_dir, subfile_name))


def subpaths(base_dir, ignore_not_exists=False):
    """Generate paths for all things in dir. Like os.path.join + listdir."""
    try:
        names = os.listdir(base_dir)
    except (IOError, OSError) as e:
        if ignore_not_exists:
            return
        raise e
    for name in names:
        yield os.path.join(base_dir, name)


def subpaths_rec(base_dir):
    for dirpath, dirnames, filenames in os.walk(base_dir, followlinks=False):
        for filename in filenames:
            yield os.path.join(dirpath, filename)


def read_kv(path):
    sep = " " * 2

    def _parse_pair(line):
        key_spec, value = line.split(sep, 1)
        key_bits = key_spec.split(":", 1)
        if len(key_bits) == 1:
            return (key_spec, value)
        key, type_ = key_bits
        assert type_ == "html"  # for now, only HTML for explicit type
        return (key, markupsafe.Markup(value))

    def _read(path):
        lines = read_file(path).strip().splitlines()
        for line in lines:
            line = line.strip()
            if line:
                yield _parse_pair(line)
    return dict(_read(path))


def read_attrs(base_dir):
    return read_kv(os.path.join(base_dir, "attrs"))


def read_translations(base_dir):
    try:
        return read_kv(os.path.join(base_dir, "translations"))
    except IOError:
        return {}


def read_kv_dir(base_dir):
    """Construct mapping {file name: file contents} from dir with files."""
    data = {}
    for file_name in os.listdir(base_dir):
        data[file_name] = read_subfile(base_dir, file_name)
    return data


def read_chunks(base_dir):
    chunks_data = {}
    # Read long chunks.
    long_chunks_dir = os.path.join(base_dir, "long_chunks")
    chunks_data.update(read_kv_dir(long_chunks_dir))
    # Read short chunks.
    short_chunks_file = os.path.join(base_dir, "short_chunks")
    chunks_data.update(read_kv(short_chunks_file))
    return chunks_data


def read_slugs(base_dir, subfile_name):
    def _read(base_dir, subfile_name):
        lines = read_subfile(base_dir, subfile_name).strip().splitlines()
        for line in lines:
            line = line.strip()
            if line:
                yield line
    return list(_read(base_dir, subfile_name))


def check_required_attrs(attrs, attr_names, entity_slug, error_callback):
    for name in attr_names:
        if name not in attrs:
            error_callback(
                "{!r} attr is not present for {!r}".format(
                    name, entity_slug))


def get_tags(note, site):
    for tag_slug, tag in site.tags.items():
        if tag_slug in note.tags_slugs:
            yield tag


def tag_sorting_key(tag):
    return tag.attrs["name"].casefold()


def sort_tags(tags):
    return sorted(tags, key=tag_sorting_key)


def sort_notes(notes):
    return sorted(notes, key=lambda note: note.date, reverse=True)


def get_substatic_data(obj_dir):
    root = os.path.join(obj_dir, "substatic")
    if os.path.isdir(root):
        suffixes = [
            os.path.relpath(subpath, start=root)
            for subpath in subpaths_rec(root)]
        return root, suffixes
    return None, []



from urllib.parse import urljoin
import datetime

import pytz
from paka.feedgenerator import Atom1Feed

from . import translations
from . import utils


_DEFAULT_TIME = datetime.time(
    hour=0, minute=0, second=0, microsecond=0, tzinfo=pytz.utc)


def make_notes_feed(template_context):
    site = template_context["site"]
    translations_prefix = template_context["view_name"]
    default_context = {"site": site}
    if template_context["is_tag_view"]:
        default_context["tag"] = template_context["tag"]
    base_url = "http://{site.attrs[domain]}".format(site=site)

    def _mkurl(path):
        return urljoin(base_url, path)

    def _mktr(suffix, **extra_context):
        return translations.translate(
            "_".join((translations_prefix, suffix)),
            context=dict(default_context, **extra_context),
            site=site, escape=False)

    try:
        subtitle = _mktr("subtitle")
    except KeyError:
        subtitle = None
    feed = Atom1Feed(
        title=_mktr("title"),
        subtitle=subtitle,
        link=_mkurl(template_context["link_path"]),
        language=site.attrs["language"],
        author_name=_mktr("author_name"),
        description=None,
        feed_url=_mkurl(template_context["url_path"]))
    for note, note_path in template_context["notes"]:
        note_url = _mkurl(note_path)
        # Notes have only date, so add time with UTC timezone.
        updateddate = datetime.datetime.combine(note.date, _DEFAULT_TIME)
        # We use slugs, as they are more stable than tag names.
        categories = [
            tag.slug
            for tag in utils.sort_tags(utils.get_tags(note, site=site))]
        feed.add_item(
            title=_mktr("item_title", note=note),
            link=note_url,
            description=note.body,
            unique_id=note_url,
            unique_id_is_permalink=True,
            updateddate=updateddate,
            categories=categories)
    return feed.writeString("utf-8")



import collections

import mako.lookup
import mako.template
from paka.webstatic.htmlmin import htmlmin

from .highlighting import render_commonmark


_TemplatePath = collections.namedtuple(
    "_TemplatePath", ["prepended", "internal", "appended"])


def make_renderer(charset, prepended, internal, appended):
    templatepath = _TemplatePath(
        prepended=list(prepended), internal=list(internal),
        appended=list(appended))
    return _Renderer(charset, templatepath)


class _Renderer(object):

    def __init__(self, charset, templatepath):
        directories = []
        for part in templatepath:
            directories.extend(part)
        self._template_lookup = mako.lookup.TemplateLookup(
            directories=directories, input_encoding=charset,
            output_encoding=charset, default_filters=["decode.utf8"],
            filesystem_checks=False, strict_undefined=True)
        self._charset = charset

    def __call__(self, template_name, **kwargs):
        template = self._template_lookup.get_template(template_name)
        return htmlmin(template.render(**kwargs).decode(self._charset))

    def render_text(self, text, **kwargs):
        return htmlmin(
            mako.template.Template(
                text, lookup=self._template_lookup).render(**kwargs))

    def render_markdown(self, text):
        return htmlmin(render_commonmark(text))



import enum
import collections

from . import errorpages
from . import features


NONE_AT_ALL = object()


_Route = collections.namedtuple(
    "_Route", [
        "url_path_tmpl", "view_name", "template_name", "fmt",
        "touch_marker", "parent_view_name"])
Fmt = enum.Enum("Fmt", {"html": "html", "atom": "atom", "direct": None})


class Map(object):

    def __init__(self, error_callback):
        self._error_callback = error_callback
        self._routes = collections.OrderedDict()

    def add_route(
            self, url_path_tmpl, view_name, template_name, fmt,
            parent_view_name):
        route = _Route(
            url_path_tmpl=url_path_tmpl, view_name=view_name,
            template_name=template_name, fmt=fmt, touch_marker=[],
            parent_view_name=parent_view_name)
        self._routes[view_name] = route
        return route

    def find_route_by_view_name(self, view_name, touch=False):
        try:
            route = self._routes[view_name]
        except KeyError:
            return
        if touch and not route.touch_marker:
            route.touch_marker.append(1)
        return route

    def check_routes(self):
        for route in self._routes.values():
            if not route.url_path_tmpl.startswith("/"):
                return self._error_callback(
                    "{!r} route does not start with slash!".format(
                        route.view_name))
            # Check parent exists.
            parent_view_name = route.parent_view_name
            if parent_view_name:
                if parent_view_name is NONE_AT_ALL:
                    continue
                found = self.find_route_by_view_name(parent_view_name)
                if not found:
                    return self._error_callback(
                        "{!r} route has non-existing parent ({!r})!".format(
                            route.view_name, parent_view_name))

    def check_for_untouched_routes(self):
        for route in self._routes.values():
            if len(route.touch_marker) < 1:
                return self._error_callback(
                    "{!r} route is not touched!".format(route.view_name))

    def format_url_path(self, view_name, context=None):
        route = self.find_route_by_view_name(view_name)
        return route.url_path_tmpl.format(**(context or {}))

    def get_lineage(self, route):
        parent_view_name = route.parent_view_name

        # Lineage concept is not applicable.
        if parent_view_name is NONE_AT_ALL:
            return []

        # Route has no parent.
        if not parent_view_name:
            return [route]

        # There is at least one parent.
        collected = []
        while parent_view_name:
            collected.append(route)
            parent_view_name = route.parent_view_name
            route = self.find_route_by_view_name(parent_view_name)
        return reversed(collected)


def add_routes(routes_map, feature_checker):
    routes_map.add_route(
        "/tags/{tag.slug}/feed/", view_name="recent_tag_notes_feed",
        template_name="notes_feed.mako", fmt=Fmt.atom,
        parent_view_name=NONE_AT_ALL)
    routes_map.add_route(
        "/tags/{tag.slug}/", view_name="one_tag",
        template_name="one_tag.html", fmt=Fmt.html,
        parent_view_name="all_tags")
    routes_map.add_route(
        "/tags/", view_name="all_tags",
        template_name="all_tags.html", fmt=Fmt.html, parent_view_name="home")
    routes_map.add_route(
        "/notes/{note.slug}/", view_name="one_note",
        template_name="one_note.html", fmt=Fmt.html,
        parent_view_name="all_notes")
    routes_map.add_route(
        "/notes/feed/", view_name="recent_notes_feed",
        template_name="notes_feed.mako", fmt=Fmt.atom,
        parent_view_name=NONE_AT_ALL)
    routes_map.add_route(
        "/notes/", view_name="all_notes",
        template_name="all_notes.html", fmt=Fmt.html, parent_view_name="home")
    if feature_checker(features.Feature.about):
        routes_map.add_route(
            "/about/", view_name="about",
            template_name="about.html", fmt=Fmt.html, parent_view_name="home")
    routes_map.add_route(
        "/", view_name="home",
        template_name="home.html", fmt=Fmt.html, parent_view_name=None)
    routes_map.add_route(
        "/robots.txt", view_name="robots_txt",
        template_name="robots.txt", fmt=Fmt.direct,
        parent_view_name=NONE_AT_ALL)
    for code in errorpages.CODES:
        view_name = errorpages.make_view_name(code)
        template_name = "{}.html".format(view_name)
        routes_map.add_route(
            errorpages.make_url_path(code), view_name=view_name,
            template_name=template_name, fmt=Fmt.html, parent_view_name=None)



import os
import shutil
import collections

from paka import breadcrumbs

from . import translations
from . import errorpages
from . import routing
from . import consts
from . import features
from . import utils


REQUIRED_CHUNK_NAMES = [
    "primary_about",
    "footer_copyrights"]
for code in errorpages.CODES:
    REQUIRED_CHUNK_NAMES.extend(
        [
            "error_page_{}_content".format(code)])


_PageSpec = collections.namedtuple(
    "_PageSpec", ["dest_path", "src_path", "contents"])


def prepare_build_dir(build_dir):
    try:
        shutil.rmtree(build_dir, ignore_errors=False)
    except (OSError, IOError):
        pass
    os.makedirs(build_dir)


def _make_chunks(site, context, required_chunk_names, error_callback):
    def _make_chunk_renderer(chunk_template, context):
        def _render_chunk():
            return site.renderer.render_text(chunk_template, **context)
        return _render_chunk
    chunks_data = dict(site.network.chunks_data, **site.chunks_data)
    chunks = {}
    for chunk_name, chunk_template in chunks_data.items():
        chunks[chunk_name] = _make_chunk_renderer(chunk_template, context)
    for chunk_name in required_chunk_names:
        if chunk_name not in chunks:
            error_callback(
                "{!r} chunk does not exist for {} site!".format(
                    chunk_name, site.slug))
    return chunks


def _sorted_series(tags):
    return sorted(tags, key=lambda tag: tag.attrs["name"])


def _make_breadcrumbs(current_route, context, site):
    routes_map = context["routes_map"]
    lineage = routes_map.get_lineage(current_route)
    if not lineage:
        return

    def _make_crumb(route):
        view_name = route.view_name

        def mk_key(suff):
            return "breadcrumbs_{}_{}".format(view_name, suff)

        label = translations.translate(
            mk_key("label"), context=context, site=site)
        try:
            heading = translations.translate(
                mk_key("heading"), context=context, site=site)
        except KeyError:
            heading = None
        return breadcrumbs.Crumb(
            label=label, heading=heading,
            url_path=routes_map.format_url_path(
                route.view_name, context=context),
            extra={"view_name": view_name})

    return breadcrumbs.Bread.from_crumbs(map(_make_crumb, lineage))


def _make_page_spec_factory(
        routes_map, site, pages_build_dir, required_chunk_names,
        extra_template_context, error_callback):
    def _make_fs_path(url_path, fmt):
        segments = [] if url_path == "/" else url_path.lstrip("/").split("/")
        if fmt is not routing.Fmt.direct:
            segments.append("index.{}".format(fmt.value))
        return os.path.join(pages_build_dir, *segments)

    def _make_page_spec(view_name, context, substatic=False):
        route = routes_map.find_route_by_view_name(view_name, touch=True)
        url_path = routes_map.format_url_path(
            route.view_name, context=context)
        if substatic:
            obj = context["substatic_object"]
            suffixes = obj.substatic_suffixes
            src_root = obj.substatic_root
            dest_root = _make_fs_path(url_path, fmt=routing.Fmt.direct)
            return (
                _PageSpec(
                    src_path=os.path.join(src_root, suffix),
                    dest_path=os.path.join(dest_root, suffix),
                    contents=None)
                for suffix in suffixes)
        else:
            # Add always-present (in templates) items.
            context = dict(
                context, view_name=view_name, url_path=url_path, site=site,
                routes_map=routes_map, **extra_template_context)
            # This statement is separate from above, because we want to avoid
            # having chunks and breadcrumbs present in context for them.
            context.update(
                breadcrumbs=_make_breadcrumbs(
                    route, context=context, site=site),
                chunks=_make_chunks(
                    site, context=context,
                    required_chunk_names=required_chunk_names,
                    error_callback=error_callback))
            # Now make page spec.
            contents = site.renderer(route.template_name, **context)
            path = _make_fs_path(url_path, fmt=route.fmt)
            return _PageSpec(dest_path=path, contents=contents, src_path=None)

    return _make_page_spec


def _make_extra_template_context(site):
    recent_notes = utils.sort_notes(site.notes)[:10]

    def popular_tags_sorting_key(tag):
        return (
            -len(tag.notes_slugs), utils.tag_sorting_key(tag))

    popular_tags = sorted(
        site.tags.values(), key=popular_tags_sorting_key)[:10]
    return {
        consts.RECENT_NOTES_TEMPLATE_CONTEXT_KEY: recent_notes,
        consts.POPULAR_TAGS_TEMPLATE_CONTEXT_KEY: popular_tags}


def _make_note_pairs_for_feed(notes, routes_map):
    for note in notes:
        note_path = routes_map.format_url_path(
            "one_note", context={"note": note})
        yield (note, note_path)


def _generate_pages_specs(
        site, pages_build_dir, required_chunk_names, feature_checker,
        error_callback):
    routes_map = routing.Map(error_callback=error_callback)
    routing.add_routes(routes_map, feature_checker=feature_checker)
    routes_map.check_routes()
    mk = _make_page_spec_factory(
        routes_map=routes_map, site=site,
        pages_build_dir=pages_build_dir,
        required_chunk_names=required_chunk_names,
        extra_template_context=_make_extra_template_context(site),
        error_callback=error_callback)
    for tag in site.tags.values():
        notes = utils.sort_notes(
            note for note in site.notes if tag.slug in note.tags_slugs)
        yield mk(
            view_name="recent_tag_notes_feed",
            context={
                "tag": tag,
                "is_tag_view": True,
                "notes": _make_note_pairs_for_feed(
                    notes[:10], routes_map=routes_map),
                "link_path": routes_map.format_url_path(
                    "one_tag", context={"tag": tag})})
        yield mk(
            view_name="one_tag",
            context={"tag": tag, "notes": notes})
    yield mk(
        view_name="all_tags",
        context={"tags": utils.sort_tags(site.tags.values())})
    for note in site.notes:
        tags = [site.tags[slug] for slug in note.tags_slugs]
        series = [site.series[slug] for slug in note.series_slugs]
        series_to_notes = {
            s.slug: utils.sort_notes([
                note for note in site.notes if s.slug in note.series_slugs])
            for s in series}
        context = {
            "note": note,
            "tags": utils.sort_tags(tags),
            "series": _sorted_series(series),
            "series_to_notes": series_to_notes}
        view_name = "one_note"
        yield mk(view_name=view_name, context=context)
        yield from mk(
            view_name=view_name, context=dict(context, substatic_object=note),
            substatic=True)
    yield mk(
        view_name="recent_notes_feed",
        context={
            "notes": _make_note_pairs_for_feed(
                utils.sort_notes(site.notes)[:10], routes_map=routes_map),
            "is_tag_view": False,
            "link_path": routes_map.format_url_path("all_notes", context={})})
    yield mk(
        view_name="all_notes",
        context={"notes": utils.sort_notes(site.notes)})
    if feature_checker(features.Feature.about):
        yield mk(view_name="about", context={})
    yield mk(view_name="home", context={})
    yield mk(view_name="robots_txt", context={})
    for code in errorpages.CODES:
        view_name = errorpages.make_view_name(code)
        yield mk(view_name=view_name, context={"code": code})
    routes_map.check_for_untouched_routes()


def build_site_by_spec(site_spec, feature_checker, error_callback):
    for page_spec in _generate_pages_specs(
            site=site_spec.site,
            pages_build_dir=site_spec.pages_build_dir,
            required_chunk_names=REQUIRED_CHUNK_NAMES,
            feature_checker=feature_checker,
            error_callback=error_callback):
        src_path = page_spec.src_path
        dest_path = page_spec.dest_path
        contents = page_spec.contents
        assert contents is not None or src_path
        os.makedirs(os.path.dirname(dest_path), exist_ok=True)
        if contents is not None:
            utils.write_file(dest_path, contents)
        else:
            shutil.copyfile(src_path, dest_path)



import enum


@enum.unique
class Feature(enum.Enum):
    about = 1


def make_feature_checker(requested_names, error_callback):
    requested = set()
    for name in requested_names:
        try:
            feature = Feature[name]
        except KeyError:
            error_callback("{!r} feature does not exist!".format(name))
        else:
            requested.add(feature)
    return lambda feature: feature in requested



import io
import re

import lxml.html
from pygments import (
    lexers as _lexers, formatters as _formatters,
    highlight as _pygments_highlight)
from paka.cmark import lowlevel as _lowlevel


_OPTS = _lowlevel.OPT_DEFAULT | _lowlevel.OPT_NOBREAKS | _lowlevel.OPT_UNSAFE


def render_commonmark(text):
    """Render CommonMark as HTML.

    Highlights code in code blocks.

    """
    text_bytes = _lowlevel.text_to_c(text)
    root = _lowlevel.parse_document(text_bytes, len(text_bytes), _OPTS)
    try:
        _substitute_code_blocks(root, _highlight)
        result = _lowlevel.text_from_c(_lowlevel.render_html(root, _OPTS))
    finally:
        _lowlevel.node_free(root)
    return result


class _HtmlFormatter(_formatters.HtmlFormatter):
    """HTML formatter that just wraps with pre and code tags."""

    def wrap(self, source, outfile):
        yield (0, "<pre><code>")
        yield from source
        yield (0, "</code></pre>")


def _highlight(fence_info, contents):
    """Highlight contents according to fence info."""
    if not fence_info:  # lexer name is not specified
        return
    if fence_info == "pycon3":
        lexer = _lexers.get_lexer_by_name("pycon", python3=True)
    else:
        lexer = _lexers.get_lexer_by_name(fence_info)
    return _pygments_highlight(contents, lexer, _HtmlFormatter())


_match_highlighting_comment = re.compile(
    r"^vx1\.highlighting: (?P<fence_info>[a-zA-Z0-9]+)$").match


def _substitute_code_blocks(root, callback):
    """Walk tree, replace code blocks with HTML returned from callback.

    In HTML blocks, highlight code in <pre> that have comment
    <!-- vx1.highlighting: ... --> as first child.

    """
    def _invoke_code_block_callback(old_node):
        old_contents = _lowlevel.text_from_c(
            _lowlevel.node_get_literal(old_node))
        fence_info = _lowlevel.text_from_c(
            _lowlevel.node_get_fence_info(old_node))
        new_contents = callback(fence_info, old_contents)
        if new_contents is None:  # can't create new contents, skip
            return
        new_node = _lowlevel.node_new(_lowlevel.NODE_HTML_BLOCK)
        try:
            assert _lowlevel.node_set_literal(
                new_node, _lowlevel.text_to_c(new_contents)) == 1
            assert _lowlevel.node_replace(old_node, new_node) == 1
        except Exception:
            _lowlevel.node_free(new_node)
            raise
        _lowlevel.node_free(old_node)

    def _invoke_html_block_callback(node):
        """Take HTML block, highlight code (if possible), set new literal."""
        def _get_html(body):
            assert body.tag == "body"
            for old_el in body.iter("pre"):
                children = list(old_el)
                first_child = children[0]
                if not isinstance(first_child, lxml.html.HtmlComment):
                    continue
                first_child_match = _match_highlighting_comment(
                    first_child.text.strip())
                if not first_child_match:
                    continue
                fence_info = first_child_match.group("fence_info")
                code = children[1].text_content()
                highlighted = callback(fence_info, code)
                new_el = lxml.html.fragment_fromstring(highlighted)
                old_el.getparent().replace(old_el, new_el)
            for el in body:
                yield lxml.html.tostring(el, encoding="unicode")

        root = lxml.html.fragment_fromstring(
            _lowlevel.text_from_c(_lowlevel.node_get_literal(node)),
            create_parent="body")
        buf = io.StringIO()
        buf.writelines(_get_html(root))
        new_literal = _lowlevel.text_to_c(buf.getvalue())
        assert _lowlevel.node_set_literal(node, new_literal) == 1

    iter_ = _lowlevel.iter_new(root)
    try:
        while True:
            event_type = _lowlevel.iter_next(iter_)
            if event_type == _lowlevel.EVENT_DONE:
                break
            elif event_type == _lowlevel.EVENT_ENTER:
                node = _lowlevel.iter_get_node(iter_)
                node_type = _lowlevel.node_get_type(node)
                if node_type == _lowlevel.NODE_CODE_BLOCK:
                    _invoke_code_block_callback(node)
                elif node_type == _lowlevel.NODE_HTML_BLOCK:
                    _invoke_html_block_callback(node)
    finally:
        _lowlevel.iter_free(iter_)



import setuptools


def _get_install_requirements():
    requirements = [
        "mako", "markupsafe", "pygments", "lxml", "paka.cmark",
        "paka.feedgenerator", "paka.webstatic", "paka.breadcrumbs"]
    return requirements


setuptools.setup(
    name="paka.vx1",
    description="Static blog generator.",
    version="3.5.0",
    packages=setuptools.find_packages(),
    install_requires=_get_install_requirements(),
    extras_require={"testing": []},
    include_package_data=True,
    namespace_packages=["paka"],
    zip_safe=False,
    url="https://github.com/PavloKapyshin/paka.vx1",
    keywords="blog generator static",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Other Audience",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython"],
    license="BSD",
    author="Pavlo Kapyshin",
    author_email="i@93z.org")
